"""
A re-implementation of several searching algorithms, including:
    - Linear Search
    - Binary Search

Many of these sort are inspired by their implementations in the CSC111
University of Toronto Course.
"""

from typing import Any, Union, Optional

import math

from structlinks.DataStructures.LinkedList.LinkedList import LinkedList
from structlinks.DataStructures.DoublyLinkedList.DoublyLinkedList import DoublyLinkedList
from structlinks.DataStructures.Graph.Graph import Graph
from structlinks.DataStructures.Queue.Queue import *
from structlinks.DataStructures.Stack.Stack import Stack


def linear_search(lst: Union[list, LinkedList, DoublyLinkedList], item: Any) -> bool:
    """Return a boolean representing whether or not the given item
    exists in the list.
    """
    # Iterate through all items in the list
    for i in lst:
        if item == i:  # If the target item is found
            return True

    # If we reach this point, the item is not in the list
    return False


def binary_search(lst: Union[list, LinkedList, DoublyLinkedList], item: Any) -> bool:
    """Return a boolean representing whether or not the given item
    exists in the list.

    Preconditions:
        - lst must be sorted
    """
    # Initialize our upper and lower bounds of the search area
    b = 0
    e = len(lst)

    # We will have verified the entire list when b >= e
    while b < e:
        # Get the middle index
        m = (e + b) // 2

        # Check if the middle item is the target
        if lst[m] == item:
            return True
        elif lst[m] > item:
            e = m
        else:
            b = m + 1

    # If we reach this point, the item is not in the list
    return False


def _edges_from_path(lst: list) -> list[tuple]:
    """Return a list of edges form the given path"""
    edges = []

    for index in range(len(lst) - 1):
        edges.append((lst[index], lst[index + 1]))

    return edges


def breadth_first_search(g: Graph, origin: Any, target: Any) -> Optional[list]:
    """Search for the smallest path between two nodes of a graph using Queues

    source: https://www.geeksforgeeks.org/print-paths-given-source-destination-using-bfs/
    """
    if origin in g.vertices:
        path = [origin]
        queue = Queue()
        queue.enqueue(path)

        while queue:

            path = queue.dequeue()
            last = path[len(path) - 1]

            if last == target:
                return _edges_from_path(path)

            for neighbour in g[last].get_neighbours():
                if neighbour not in path:
                    new_path = path.copy()
                    new_path.append(neighbour)
                    queue.enqueue(new_path)

        return None
    else:
        return None


def depth_first_search(g: Graph, origin: Any, target: Any) -> Optional[list]:
    """Search for the shortest path between two nodes of a graph using Stacks
    """
    if origin in g.vertices:
        path = [origin]
        stack = Stack()
        stack.push(path)

        while stack:

            path = stack.pop()
            last = path[len(path) - 1]

            if last == target:
                return _edges_from_path(path)

            for neighbour in g[last].get_neighbours():
                if neighbour not in path:
                    new_path = path.copy()
                    new_path.append(neighbour)
                    stack.push(new_path)

        return None
    else:
        return None


def dijkstra_search_target(g: Graph, origin: Any, target: Any, metric: Any) -> Optional[list]:
    """Use Dijkstra Algorithm to find the shortest path for a weighted graph from one point to
    another

    sources:
    - https://www.techiedelight.com/single-source-shortest-paths-dijkstras-algorithm/

    - https://www.freecodecamp.org/news/dijkstras-shortest-path-algorithm-visual-introduction/#:~
    :text=Dijkstra's%20Algorithm%20finds%20the%20shortest,node%20and%20all%20other%20nodes.

    - https://www.geeksforgeeks.org/printing-paths-dijkstras-shortest-path-algorithm/
    """

    path = []  # list of all the visited vertices
    prev = {item: None for item in g.vertices}  # parent-child search dict

    dist = 'distance'  # distance attribute for a vertices

    # set distance attribute for all vertices to be infinity
    g.add_global_attributes({dist: math.inf})
    # setting distance attribute 0 for the origin vertex
    g[origin].attr[dist] = 0

    # creating a function by which the queue would prioritize
    def queue_metric(value: Any) -> Any:
        """Return the metric of the vertex with value <value>"""
        return - g[value].attr[dist]

    # creating a priority queue with -distance as the metric
    queue = Queue(list(g.vertices), queue_metric)

    while not queue.is_empty:  # while the queue still have vertices un-explored
        item = queue.dequeue()  # taking out the vertex with the shortest distance
        path.append(item)  # add the item to visited

        if item == target:  # Base case : if we reached the item
            try:
                return _edges_from_path(_path_from_dict(prev, origin, target))
            except KeyError:
                return None

        for neighbour in g[item].get_neighbours():
            # if neighbour is in the queue, that us when the distance of neighbour is infinity
            if g[neighbour].attr[dist] > g[item].attr[dist] + g[item][neighbour][metric]:
                #  set the new distance
                g[neighbour].attr[dist] = g[item].attr[dist] + g[item][neighbour][metric]
                #  mark the parent item
                prev[neighbour] = item

        #  update the order of the queue
        queue.update()

    # Return None if there is no path
    return None


def dijkstra_search_all(g: Graph, origin: Any, metric: Any) -> dict:
    """Use Dijkstra Algorithm to find the shortest path for a weighted graph from one point to
    all the the other points

    sources:
    - https://www.techiedelight.com/single-source-shortest-paths-dijkstras-algorithm/

    - https://www.freecodecamp.org/news/dijkstras-shortest-path-algorithm-visual-introduction/#:~
    :text=Dijkstra's%20Algorithm%20finds%20the%20shortest,node%20and%20all%20other%20nodes.

    - https://www.geeksforgeeks.org/printing-paths-dijkstras-shortest-path-algorithm/
    """

    path = []  # list of all the visited vertices
    prev = {item: None for item in g.vertices}  # parent-child search dict

    dist = 'distance'  # distance attribute for a vertices

    # set distance attribute for all vertices to be infinity
    g.add_global_attributes({dist: math.inf})
    # setting distance attribute 0 for the origin vertex
    g[origin].attr[dist] = 0

    # creating a function by which the queue would prioritize
    def queue_metric(value: Any) -> Any:
        """Return the metric of the vertex with value <value>"""
        return - g[value].attr[dist]

    # creating a priority queue with -distance as the metric
    queue = Queue(list(g.vertices), queue_metric)

    while not queue.is_empty:  # while the queue still have vertices un-explored
        item = queue.dequeue()  # taking out the vertex with the shortest distance
        path.append(item)  # add the item to visited

        for neighbour in g[item].get_neighbours():
            # if neighbour is in the queue, that us when the distance of neighbour is infinity
            if g[neighbour].attr[dist] > g[item].attr[dist] + g[item][neighbour][metric]:
                #  set the new distance
                g[neighbour].attr[dist] = g[item].attr[dist] + g[item][neighbour][metric]
                #  mark the parent item
                prev[neighbour] = item

        #  update the order of the queue
        queue.update()

    paths_dict = {}

    for item in g.vertices:
        if item != origin:
            try:
                paths_dict[item] = _edges_from_path(_path_from_dict(prev, origin, item))
            except KeyError:
                paths_dict[item] = None

    return paths_dict


def _path_from_dict(d: dict, src: Any, target: Any) -> list:
    """Helper function for dijkstra_search"""
    current_key = target
    path = LinkedList([target])

    while current_key != src:

        path.insert(0, d[current_key])
        current_key = d[current_key]

    return path.to_list()



"""This file contains the Matrix object."""

from __future__ import annotations
from typing import Union, Callable
import copy


class Matrix:
    """A matrix

    Instance Attributes:
    - shape: shape of the matrix

    Private Instance Attributes:
    - _row_count: The number of rows
    - _col_count: The number of columns
    - _matrix: A 2-D list of integers or floats
    """
    _matrix: list[list[Union[int, float]]]
    _row_count: int
    _col_count: int

    def __init__(self, array: list[list[Union[int, float]]]) -> None:
        """Initialize a new matrix
        """
        self._row_count = len(array)
        self._col_count = len(array[0]) if self._row_count > 0 else 0

        self.shape = (self._row_count, self._col_count)

        self._matrix = array

    def _get_indices(self, s: slice, stop_num: int) -> list:
        start = s.start if s.start else 0
        end = s.stop if s.stop else stop_num
        return list(range(start, end))
    
    def __getitem__(self, index) -> list:
        """ Return the row at index <index>"""
        if isinstance(index, int):
            raise ValueError("Please provide both Row and Col in the format [row (int/slice), col (int/slice)]")
        
        elif isinstance(index[0], int) and isinstance(index[1], int):
            if index[0] >= self._row_count or index[1]>= self._col_count:
                raise IndexError("either row or column is out of bound")
            return self._matrix[index[0]][index[1]]
        
        else:
            new_matrix = []
            rows = [index[0]] if isinstance(index[0], int) else self._get_indices(index[0], self._row_count)
            cols = [index[1]] if isinstance(index[1], int) else self._get_indices(index[1], self._col_count)
            
            if any(element < 0 or element >=self._row_count for element in rows):
                if any(element < 0 or element >=self._col_count for element in self._col_count):
                    raise IndexError("either row or column is out of bound")
                
            for row in rows:
                temp = []
                for col in cols:
                    temp.append(self._matrix[row][col])
                new_matrix.append(temp)

            return Matrix(new_matrix)

    def __setitem__(self, index, value) -> None:
        """Set the value <value> of item at index <index> in self"""
        if isinstance(index, int):
            raise ValueError("Please provide both Row and Col in the format [row (int/slice), col (int/slice)]")
        
        elif isinstance(index[0], int) and isinstance(index[1], int):
            if index[0] >= self._row_count or index[1]>= self._col_count:
                raise IndexError("either row or column is out of bound")
            self._matrix[index[0]][index[1]] = value
        
        elif isinstance(index[0], slice) or isinstance(index[1], slice):
            rows = [index[0]] if isinstance(index[0], int) else self._get_indices(index[0], self._row_count)
            cols = [index[1]] if isinstance(index[1], int) else self._get_indices(index[1], self._col_count)
            
            if any(element < 0 or element >=self._row_count for element in rows):
                if any(element < 0 or element >=self._col_count for element in self._col_count):
                    raise IndexError("either row or column is out of bound")
                
            for row_index, row in enumerate(rows):
                for col_index, col in enumerate(cols):
                    self._matrix[row][col] = value[row_index, col_index]
        else:
            raise ValueError("The row and the column should be of int or slice data type")    
        
    def __str__(self) -> str:
        return '[' + '\n '.join([str(row) for row in self._matrix]) + ']'

    def __repr__(self) -> str:
        return 'Matrix([' + '\n        '.join([str(row) for row in self._matrix]) + '])'

    def copy(self) -> Matrix:
        return Matrix(copy.deepcopy(self._matrix))

    def to_list(self) -> list[list[Union[int, float]]]:
        return copy.deepcopy(self._matrix)

    @property
    def is_square(self) -> bool:
        """Return whether the current matrix is square"""
        return self._row_count == self._col_count

    @property
    def is_invertible(self) -> bool:
        """Return whether the current matrix is invertible"""
        if self.is_square:
            return self.determinant != 0
        return False

    @property
    def is_vector(self) -> bool:
        """Return whether self is a vector"""
        return self._row_count == 1 or self._col_count == 1

    @staticmethod
    def zeros(shape: tuple[int, int]) -> Matrix:
        """Return a zero matrix of shape <shape>"""
        return Matrix([[0] * shape[1] for _ in range(shape[0])])

    @staticmethod
    def ones(shape: tuple[int, int]) -> Matrix:
        """Return a matrix with all elements one of shape <shape>"""
        return Matrix([[1] * shape[1] for _ in range(shape[0])])

    @staticmethod
    def identity(order: int) -> Matrix:
        """Return a identity matrix of shape order X order"""
        return Matrix([[1 if i == j else 0 for j in range(order) for i in range(order)]])

    def map(self, func: Callable) -> Matrix:
        """Return a new matrix mapped to function <func>"""
        copy_mat = self.copy()

        for i in range(self._row_count):
            for j in range(self._col_count):
                copy_mat[i, j] = func(copy_mat[i, j])

        return copy_mat

    def add_scalar(self, k: Union[int, float]) -> Matrix:
        """Add the current matrix by the scalar k
        to produce a new matrix. This operation does not mutate
        the original matrix
        """
        return self.map(lambda x: x + k)

    def multiply_scalar(self, k: Union[int, float]) -> Matrix:
        """Multiply the current matrix by the scalar k
        to produce a new matrix. This operation does not mutate
        the original matrix
        """
        return self.map(lambda x: x * k)
    
    def add_matrix(self, other: Matrix) -> Matrix:
        """Add the current matrix and matrix b to produce a
        new matrix. This operation does not mutate either matrix

        Preconditions:
            - self.row_count == b.row_count
            - self.col_count == b.col_count
        """
        if self.shape != other.shape:
            raise ShapeError

        sum_matrix = Matrix.zeros((self._row_count, self._col_count))
        for i in range(self._row_count):
            for j in range(self._col_count):
                # Add the corresponding entries for both matrices
                sum_matrix[i, j] = self[i, j] + other[i, j]
        return sum_matrix

    def multiply_matrix(self, other: Matrix) -> Matrix:
        """Matrix multiply self with other.
        This operation does not mutate the original matrix"""
        # check if self and other can be multiplied
        if self.shape[1] != other.shape[0]:
            raise TypeError(f'Can not multiply matrices of shape {self.shape} and {other.shape}')
        
        result = Matrix.zeros((self.shape[0], other.shape[1]))
        # iterate through rows of self
        for i in range(self._row_count):
            # iterate through columns of other
            for j in range(other.shape[1]):
                # iterate through rows of outer
                for k in range(other.shape[0]):
                    result[i, j] += self[i, k] * other[k, j]

        return result
    
    @property
    def transpose(self) -> Matrix:
        """Return transposed matrix of self"""
        new_matrix = Matrix.zeros((self.shape[1], self.shape[0]))

        for i in range(self._row_count):
            for j in range(self._col_count):
                new_matrix[j, i] = copy.deepcopy(self[i, j])

        return new_matrix

    def dot_multiply(self, other: Matrix) -> Matrix:
        """Return the dot product of two matrices

        Precondition:
        - self.shape == other.shape
        """
        if self.shape != other.shape:
            raise ShapeError

        return self.multiply_matrix(other.transpose)

    def minor(self, row: int, col: int) -> Matrix:
        """Return the minor of the matrix with the given row and
        col removed. This operation does not mutate the original matrix

        Preconditions:
            - 0 <= row < self.row_count
            - 0 <= col < self.col_count 
        """
        if not(0 <= row < self._row_count and 0 <= col < self._col_count):
            raise RowColOutOfBound

        # create a new array
        copy_mat = self.to_list()
        # remove the row
        copy_mat.pop(row)
        # remove the col from all the other rows
        for row in copy_mat:
            row.pop(col)
        # return the new matrix
        return Matrix(copy_mat)

    @property
    def determinant(self) -> Union[float, int]:
        """Return the determinant of the current matrix
        
        Preconditions:
            - self.row_count == self.col_count
        """
        if not self.is_square:
            return None

        if self._row_count == 0 and self._col_count == 0:
            return 1
        elif self._row_count == 1 and self._col_count == 1:
            return self[0, 0]
        else:
            determinant = 0
            for j in range(self._col_count):
                determinant += self[0, j] * ((-1) ** j) * self.minor(0, j).determinant
            return determinant

    @property
    def cofactor(self) -> Matrix:
        """Return the cofactor matrix of self"""
        if not self.is_square:
            return None

        new_matrix = Matrix.zeros(self.shape)

        for i in range(self._row_count):
            for j in range(self._col_count):
                new_matrix[i, j] = ((-1) ** (i + j)) * self.minor(i, j).determinant

        return new_matrix

    @property
    def adjacent(self) -> Matrix:
        """Return the adj matrix of self"""
        if not self.is_square:
            return None

        return self.cofactor.transpose

    @property
    def inverse(self) -> Matrix:
        """Return inverse of self"""
        if not self.is_invertible:
            return None

        return self.adjacent.multiply_scalar(1 / self.determinant)

    @staticmethod
    def vstack(matrices: list[Matrix]) -> Matrix:
        new_matrix = []
        def_shape = None
        
        if len(matrices) == 0:
            raise ValueError("the list of matrices can not be empty")
        
        for matrix in matrices:
            
            if not def_shape:
                def_shape = matrix.shape
                
            if not def_shape[1] == matrix.shape[1]:
                raise ColumnError
            
            for row in matrix.to_list():
                new_matrix.append(row)
                
        return Matrix(new_matrix)
    
    @staticmethod
    def hstack(matrices: list[Matrix]) -> Matrix:
        new_matrix = []
        def_shape = None
        
        if len(matrices) == 0:
            raise ValueError("the list of matrices can not be empty")
        
        for matrix in matrices:
            
            if not def_shape:
                def_shape = matrix.shape
                new_matrix = matrix.to_list()
            
            elif not def_shape[0] == matrix.shape[0]:
                raise RowError
            
            else:
                for row_index in range(def_shape[0]):
                    mat_row = matrix._matrix[row_index]
                    new_matrix[row_index].extend(mat_row) 
                                
        return Matrix(new_matrix)
    
    @staticmethod
    def _ToReducedRowEchelonForm( M):
        if not M: return
        lead = 0
        rowCount = len(M)
        columnCount = len(M[0])
        for r in range(rowCount):
            if lead >= columnCount:
                return
            i = r
            while M[i][lead] == 0:
                i += 1
                if i == rowCount:
                    i = r
                    lead += 1
                    if columnCount == lead:
                        return
            M[i],M[r] = M[r],M[i]
            lv = M[r][lead]
            M[r] = [ mrx / float(lv) for mrx in M[r]]
            for i in range(rowCount):
                if i != r:
                    lv = M[i][lead]
                    M[i] = [ iv - lv*rv for rv,iv in zip(M[r],M[i])]
            lead += 1 
               
    @property
    def rref(self) -> Matrix:
        """Return the rref form of self
        
        source: https://rosettacode.org/wiki/Reduced_row_echelon_form#Python
        """
        M = self.to_list()
        Matrix._ToReducedRowEchelonForm(M)
        return Matrix(M)
        
    @property
    def rank(self) -> int:
        rank = 0
        
        for row in self.rref._matrix:
            if any(element != 0 for element in row):
                rank += 1
        
        return rank   
    
    @property
    def nullity(self) -> int:
        return self._col_count - self.rank
    
    @property
    def linearly_independent(self) -> bool:
         return self.nullity == 0    
    
    @property
    def linearly_dependent(self) -> bool:
         return self.nullity != 0
    
    def get_independent_vectors(self) -> list[Matrix]:
            vectors = []
            
            for index, row in enumerate(self.rref._matrix):
                if any(element != 0 for element in row):
                    vectors.append(self[index, :])
            
            return vectors        
    
    def get_dependent_vectors(self) -> list[Matrix]:
            vectors = []
            
            for index, row in enumerate(self.rref._matrix):
                if all(element == 0 for element in row):
                    vectors.append(self[index, :])
            
            return vectors  
        
        
class NotSquareMatrix(Exception):
    """Error for when a non square matrix is passed"""
    def __str__(self) -> str:
        return "The matrix needs to be a square matrix"


class RowColOutOfBound(Exception):
    """Error for when the row or col index is out of bound"""
    def __str__(self) -> str:
        return "Either or Both the Row or Col index are out of bound"


class ShapeError(Exception):
    """Error for when the shape of the two matrices is different"""
    def __str__(self) -> str:
        return "The shape of the matrices needs to be the same"


class InverseDoesNotExists(Exception):
    """Error for when inverse does not exists"""
    def __str__(self) -> str:
        return "The matrix is not invertible"


class RowError(Exception):
    """Error for when the number of rows are not the same"""
    def __str__(self) -> str:
        return "The number of rows of the matrices need to be the same"


class ColumnError(Exception):
    """Error for when the number of columns are not the same"""
    def __str__(self) -> str:
        return "The number of columns of the matrices need to be the same"
       


"""
This is an implementation of the stack data structure, with code pulled from the
University of Toronto's CSC110 course notes.
"""
from __future__ import annotations
from typing import Any, Optional, Callable, Sequence


class Stack:
    """
    This class represents a stack data structure
    """

    def __init__(self, items: Optional[list] = None) -> None:
        """
        Initialize a stack, empty at first
        """
        self.items = []

        if items:
            self.items = items

    def is_empty(self) -> bool:
        """
        Returns whether the stack is empty
        """
        return self.items == []

    def push(self, item: Any) -> None:
        """
        Adds a new element to the top of the stack
        """
        self.items.append(item)

    def push_multiple(self, items: Sequence) -> None:
        """Push multiple items in the stack"""
        for item in items:
            self.push(item)

    def pop(self) -> Any:
        """
        Removes the element at the top of the stack and
        returns it. Raises IndexError if list is empty
        """
        if self.is_empty():
            raise EmptyStackError
        else:
            return self.items.pop()

    def __len__(self) -> int:
        """Return the length of the stack"""
        return len(self.items)

    def to_list(self) -> list:
        """Return list of the stack"""
        return self.items

    def __copy__(self) -> Stack:
        """Return a copy of the stack"""
        return Stack(items=self.items)

    def copy(self) -> Stack:
        """Return a copy of the stack"""
        return Stack(items=self.items)

    def map(self, key: Callable) -> Stack:
        """Map a function to the stack"""
        return Stack(items=[key(item) for item in self.items])

    def invert(self) -> None:
        """Invert the stack"""
        self.items.reverse()

    def extend(self, other: Stack):
        """extend the stack by putting other stack on top of self"""
        self.push_multiple(other.items)

    def __add__(self, other) -> Stack:
        """Return a stack with other stack on top of self"""
        return Stack(items=self.items + other.items)

    def __str__(self) -> str:
        """Return string representation of the stack"""
        string_so_far = ''

        gap = 10

        for index in range(len(self) - 1, -1, -1):
            item = self.items[index]

            string_rep = str(item)

            gap_left = gap - len(string_rep)

            string_so_far += '|' + ' ' * gap + string_rep + ' ' * gap_left + '| \n'

        string_so_far += '|' + '_' * (2 * gap) + '|'

        return string_so_far

    def __repr__(self) -> str:
        return f'Stack({self.items})'


class EmptyStackError(Exception):
    def __str__(self) -> str:
        """String representation of the error"""
        return "Popping from an empty stack is not allowed"



"""The file contains Graph object. The base code for the vertex and the graph object
is taken from University of Toronto's CSC111 course."""

from __future__ import annotations
from typing import Any, Optional, Union, Callable


class _Vertex:
    """A vertex in a graph.

    Instance Attributes:
        - item: The data stored in this vertex.
        - attributes: attributes of the vertex
        - neighbours: The vertices that are adjacent to this vertex.
    """
    item: Any
    attributes: dict
    neighbours: dict

    def __init__(self, item: Any,
                 neighbours: Optional[dict] = None,
                 attributes: Optional[dict] = None) -> None:
        """Initialize a new vertex with the given item and neighbours."""
        self.item = item
        self.attributes = attributes if attributes else {}
        self.neighbours = neighbours if neighbours else {}
        self.view = _VertexView(self)

    def check_connected(self, target_item: Any, visited: set[_Vertex]) -> bool:
        """Return whether this vertex is connected to a vertex corresponding to the target_item,
        WITHOUT using any of the vertices in visited.

        Preconditions:
            - self not in visited

        The base code is taken from University of Toronto's CSC111 course
        """
        if self.item == target_item:
            # Our base case: the target_item is the current vertex
            return True
        else:
            visited.add(self)  # Add self to the set of visited vertices
            for u in self.neighbours:
                u_vertex = self.neighbours[u]['item']
                if u_vertex not in visited:  # Only recurse on vertices that haven't been visited
                    if u_vertex.check_connected(target_item, visited):
                        return True

            return False

    def get_connected(self, visited: set[_Vertex]) -> set:
        """Return a set of all ITEMS connected to self by a path that does not use
        any vertices in visited.

        The items of the vertices in visited CANNOT appear in the returned set.

        Preconditions:
            - self not in visited

        The base code is taken from University of Toronto's CSC111 course
        """

        all_in_visited = True

        items_set = {self.item}

        for neighbour in self.neighbours:

            neighbour_vertex = self.neighbours[neighbour]['item']

            if neighbour_vertex not in visited:
                all_in_visited = False

        if all_in_visited:
            visited.add(self)
            return items_set

        else:
            visited.add(self)

            for neighbour in self.neighbours:

                neighbour_vertex = self.neighbours[neighbour]['item']

                if neighbour_vertex not in visited:
                    items_set = items_set.union(neighbour_vertex.get_connected(visited))

            return items_set

    def spanning_graph(self, visited: set[_Vertex]) -> list[tuple]:
        """Return a Graph that form a spanning tree of all vertices that are
        connected to this vertex WITHOUT using any of the vertices in visited.

        The edges are returned as a list of sets, where each set contains the two
        ITEMS corresponding to an edge.

        Preconditions:
            - self not in visited

        The base code is taken from University of Toronto's CSC111 course
        """
        edges_so_far = []

        visited.add(self)

        for neighbour in self.neighbours:
            # Only recurse on vertices that haven't been visited
            if self.neighbours[neighbour]['item'] not in visited:
                edges_so_far.append((self.item, neighbour,
                                     self.neighbours[neighbour]['attributes']))
                edges_so_far.extend(self.neighbours[neighbour]['item'].spanning_graph(visited))

        return edges_so_far

    def paths_to(self, item: Any, visited: set, current_path: list, paths: set) -> None:
        """Return a set of all the paths between self and item"""

        visited.add(self.item)
        current_path.append(self.item)

        if self.item == item:
            paths.add(tuple(current_path))

        else:
            for neighbour in self.neighbours:
                if neighbour not in visited:
                    self.neighbours[neighbour]['item'].paths_to(item, visited, current_path, paths)

        current_path.pop()
        visited.remove(self.item)

    def update_attributes(self, other) -> None:
        """update the attributes of the edge"""
        if other in self.neighbours:

            other_vertex = self.neighbours[other]['item']

            attributes = other_vertex.neighbours[self.item]['attributes']

            self.neighbours[other]['attributes'].update(attributes)


class _VertexView:
    """Used to get/view a vertex"""

    def __init__(self, _vertex: _Vertex) -> None:
        self._vertex = _vertex

    def __repr__(self) -> str:
        """Return string representation of vertex"""

        string_so_far = 'NeighboursView({ \n'

        for neighbour in self._vertex.neighbours:
            string_so_far += \
                '  {' + f'{neighbour}: {self._vertex.neighbours[neighbour]["attributes"]}' + '}, \n'

        return string_so_far + '})'

    def __setitem__(self, other: Any, properties: dict) -> None:
        """Set a property of the edge between self and other

        properties is of the form:
        - {'property 1': value 1, 'property 2': value 2, ... 'property n': value n}
        """
        if other not in self._vertex.neighbours:
            raise ValueError('The vertex value is not a neighbour to self, or does not exists')

        if not isinstance(properties, dict):
            raise TypeError('attribute properties should be a dictionary')

        self._vertex.neighbours[other]['attributes'] = properties

    def __getitem__(self, other: Any) -> dict:
        """Return all the attributes of edge between self and other"""
        if other not in self._vertex.neighbours:
            return {}

        other_vertex = self._vertex.neighbours[other]['item']

        self._vertex.update_attributes(other)
        other_vertex.update_attributes(self._vertex.item)

        return self._vertex.neighbours[other]['attributes']

    @property
    def attr(self) -> dict:
        """Return the attribute dict of the vertex"""
        return self._vertex.attributes

    def check_connected(self, item: Any) -> bool:
        """Check if vertex is connected to the item"""
        return self._vertex.check_connected(item, set())

    def get_connected(self) -> set:
        """Return set of all elements connected to vertex"""
        return self._vertex.get_connected(set())

    def in_cycle(self) -> bool:
        """Return whether vertex is in a cycle or not

        The base code is taken from University of Toronto's CSC111 course
        """

        return any(self._vertex.neighbours[neighbour1]['item'].check_connected(neighbour2,
                                                                               {self._vertex})
                   for neighbour1 in self._vertex.neighbours
                   for neighbour2 in self._vertex.neighbours
                   if neighbour1 != neighbour2)

    def adjacent(self, item: Any) -> bool:
        """Return whether item is adjacent to self.
        """
        return item in self._vertex.neighbours

    def get_neighbours(self) -> set:
        """Return a set of the neighbours of vertex
        """
        return set(self._vertex.neighbours.keys())

    @property
    def degree(self) -> int:
        """Return the degree of the vertex"""
        return len(self._vertex.neighbours)

    def create_spanning_graph(self) -> Graph:
        """Create a spanning graph from this vertex"""
        graph = Graph()
        edges = self._vertex.spanning_graph(set())
        graph.add_edges(edges)
        return graph

    def get_all_paths_to(self, item: Any) -> set:
        """Return all the paths to the item from self"""
        paths = set()
        self._vertex.paths_to(item, set(), [], paths)
        return paths


class Graph:
    """A graph.

    The base code for this class taken from University of Toronto's CSC111 course
    """
    # Private Instance Attributes:
    #     - _vertices: A collection of the vertices contained in this graph.
    #                  Maps item to _Vertex instance.
    _vertices: dict[Any, _Vertex]

    def __init__(self) -> None:
        """Initialize an empty graph (no vertices or edges)."""
        self._vertices = {}

    def add_vertex(self, item: Any, attributes: Optional[dict] = None) -> None:
        """Add a vertex with the given item to this graph.

        The new vertex is not adjacent to any other vertices.

        The base code is taken from University of Toronto's CSC111 course
        """
        if item not in self._vertices:
            self._vertices[item] = _Vertex(item, attributes=attributes if attributes else {})
        else:
            self._vertices[item].attributes.update(attributes if attributes else {})

    def add_vertices(self, items: list[Any], global_attributes: Optional[dict] = None):
        """Add multiple vertices

        An items can look like:
        - [item 1, item 2, ..., item n]
        - [(item 1, attribute 1), (item 2, attribute 2), ..., (item n, attribute n)]
        - [(item 1, attribute 1), item2, item3, ..., (item n, attribute n)]

        global_attributes will be added to all the vertices

        The base code is taken from University of Toronto's CSC111 course
        """
        if not global_attributes:
            global_attributes = {}

        for item in items:

            if isinstance(item, tuple) and len(item) == 2:
                item_value, item_attributes = item
                item_attributes.update(global_attributes)
                self.add_vertex(item_value, attributes=item_attributes)
            else:
                self.add_vertex(item)

    @property
    def vertices(self) -> set:
        """Return set of all the vertices in the graph"""
        return set(self._vertices.keys())

    def add_edge(self, item1: Any, item2: Any, attributes: Optional[dict] = None) -> None:
        """Add an edge between the two vertices with the given items in this graph,
        and with the give attributes. attributes here is the list of properties if
        the edge between the two items.

        The base code is taken from University of Toronto's CSC111 course
        """
        # check if item1 != item2
        if item1 != item2:
            # if item1 is not in self._vertices then add it
            if item1 not in self._vertices:
                self.add_vertex(item1)
            # if item2 is not in self._vertices then add it
            if item2 not in self._vertices:
                self.add_vertex(item2)
            # extract the _Vertex object of item1 and item2
            v1 = self._vertices[item1]
            v2 = self._vertices[item2]

            if item1 not in v2.neighbours and item2 not in v1.neighbours:
                # Add the new edge
                v1.neighbours[item2] = {'item': v2, 'attributes': attributes if attributes else {}}
                v2.neighbours[item1] = {'item': v1, 'attributes': attributes if attributes else {}}
            else:
                v1.neighbours[item2]['attributes'].update(attributes if attributes else {})
                v2.neighbours[item1]['attributes'].update(attributes if attributes else {})

    def add_edges(self, edges: Union[list[tuple[Any, Any]], list[tuple[Any, Any, dict]]]) -> None:
        """Add multiple edges to the graph

        Instance Attributes:
        - edges: it is a list of tuples representing an edge. The tuple could be of length 2 or 3

                if edge is of length 2 then it would look like:
                - (item1, item2)

                if edge is of length 2 then it would look like:
                - (item1, item2, edge-attributes)

        The base code is taken from University of Toronto's CSC111 course
        """
        for edge in edges:
            if not (len(edge) == 2 or len(edge) == 3):
                raise TypeError('The edges are not formatted properly')

        for edge in edges:

            if len(edge) == 2:
                self.add_edge(edge[0], edge[1])

            else:
                self.add_edge(edge[0], edge[1], edge[2])

    @property
    def edges(self) -> set:
        """Return set of all the edges"""

        # Set of all the edges
        set_so_far = set()
        # set of all the elements which have been visited
        visited = set()

        for vertex in self._vertices:
            # add vertex to visited
            visited.add(vertex)

            for neighbour in self._vertices[vertex].neighbours:
                # only add edge if not previously added
                if neighbour not in visited:
                    set_so_far.add((vertex, neighbour))

        return set_so_far

    def clear(self) -> None:
        """clear the graph"""
        for vertex in self.vertices:
            del self._vertices[vertex]
        self._vertices = {}

    def adjacent(self, item1: Any, item2: Any) -> bool:
        """Return whether item1 and item2 are adjacent vertices in this graph.

        Return False if item1 or item2 do not appear as vertices in this graph.

        The base code is taken from University of Toronto's CSC111 course
        """
        if item1 in self._vertices and item2 in self._vertices:
            return item2 in self._vertices[item1].neighbours
        else:
            # We didn't find an existing vertex for both items.
            return False

    def get_neighbours(self, item: Any) -> set:
        """Return a set of the neighbours of the given item.

        Raise a ValueError if item does not appear as a vertex in this graph.

        The base code is taken from University of Toronto's CSC111 course
        """
        if item in self._vertices:
            v = self._vertices[item]
            return set(v.neighbours.keys())
        else:
            raise ValueError

    @property
    def vertices_count(self) -> int:
        """Return the number of vertices"""
        return len(self._vertices)

    @property
    def edges_count(self) -> int:
        """Return the number of edges"""
        return len(self.edges)

    def get_degree(self, item: Union[set, Any]) -> Union[dict, int]:
        """Return the degree of the item in the graph if item is a single element
        otherwise if item is a set of elements then it would return a dict of
        elements mapped to their degrees

        Raise a ValueError if item does not appear as a vertex in this graph.
        """
        # If item is a set of elements
        if isinstance(item, set):
            dict_so_far = {}

            for element in item:
                dict_so_far[element] = self.get_degree(element)

            return dict_so_far
        # if item is a single element
        if item in self._vertices:
            return self[item].degree
        else:
            raise ValueError('The element or one of the elements is not on the graph')

    def __getitem__(self, item) -> _VertexView:
        """Returns a object containing all the information about the neighbours of vertex,
        and used to set/get properties of edges
        """
        if item not in self._vertices:
            raise ValueError

        return self._vertices[item].view

    def remove_vertex(self, item: Any) -> None:
        """Delete item from the graph"""
        if item in self._vertices:

            for neighbour in self._vertices[item].neighbours:
                self._vertices[neighbour].neighbours.pop(item)

            self._vertices.pop(item)

        else:
            raise ValueError('The item is not in the graph')

    def remove_edge(self, item1: Any, item2: Any) -> None:
        """Remove edge between item1 and item2"""
        if item1 in self._vertices and item2 in self._vertices:
            if item1 in self.get_neighbours(item2):
                self._vertices[item1].neighbours.pop(item2)
                self._vertices[item2].neighbours.pop(item1)
        else:
            raise ValueError('One/both of item1, item2 is/are not in the graph')

    def remove_vertices(self, items: list) -> None:
        """Remove all the vertices in items"""
        if not all(item in self._vertices for item in items):
            raise ValueError('One or more items in the list are not present in the graph')

        else:
            for item in items:
                self.remove_vertex(item)

    def remove_edges(self, edges: list[tuple]) -> None:
        """Remove all the edge pairs in <edges>"""
        if not all(item1 in self._vertices and item2 in self._vertices
                   for (item1, item2) in edges):
            raise ValueError('One or more of the edge pairs have elements which are not in graph')

        for (item1, item2) in edges:
            self.remove_edge(item1, item2)

    def add_global_attributes(self, attributes: dict) -> None:
        """Add attributes to all the vertices in the graph"""
        for item in self.vertices:
            self[item].attr.update(attributes)

    def sub_graph_by_value(self, predicate: Callable) -> Graph:
        """Return a sub-graph of self, with vertices which pass the predicate function"""
        subgraph = Graph()

        subgraph.add_vertices([(item, self[item].attr)
                               for item in self.vertices if predicate(item)])

        vertices = subgraph.vertices

        for (item1, item2) in self.edges:
            if item1 in vertices and item2 in vertices:
                subgraph.add_edge(item1, item2, self[item1][item2].copy())

        return subgraph

    def sub_graph_by_attribute(self, attribute: Any,
                               predicate: Optional[Callable] = lambda x: True) -> Graph:
        """Return a sub-graph of self, with vertices which contain attribute and
         pass the predicate function for that attribute"""

        subgraph = Graph()

        subgraph.add_vertices([(item, self[item].attr)
                               for item in self.vertices
                               if attribute in self[item].attr and predicate(item)])

        vertices = subgraph.vertices

        for (item1, item2) in self.edges:
            if item1 in vertices and item2 in vertices:
                subgraph.add_edge(item1, item2, self[item1][item2].copy())

        return subgraph

    def union(self, other: Graph) -> Graph:
        """Return a new graph which contains all the vertices of self and other. Attributes
        for duplicates would be combined to also form a union."""

        graph = Graph()

        graph.add_vertices([(item, self[item].attr) for item in self.vertices])
        graph.add_vertices([(item, other[item].attr) for item in other.vertices])

        graph.add_edges([(item1, item2, self[item1][item2].copy())
                         for (item1, item2) in self.edges])

        graph.add_edges([(item1, item2, other[item1][item2].copy())
                         for (item1, item2) in other.edges])

        return graph

    @property
    def is_connected(self) -> bool:
        return all(self[item1].check_connected(item2)
                   for item1 in self.vertices for item2 in self.vertices)

    def create_spanning_graph(self) -> Graph:
        """Return a spanning graph of self

        Precondition:
        - self.is_connected

        The base code is taken from University of Toronto's CSC111 course
        """
        if not self.is_connected:
            raise TypeError('The graph should be connected')

        if self.vertices_count > 0:
            return self[list(self.vertices)[0]].create_spanning_graph()

        else:
            return Graph()

    def __str__(self) -> str:
        """Return a string representation of the graph"""
        string_representation = '{ \n'

        for item in self.vertices:
            string_representation += f'{item}: {self[item].get_neighbours()} \n'

        string_representation += '}'

        return string_representation

    def __repr__(self) -> str:
        """Return a string representation of the graph"""
        string_representation = '{ \n'

        for item in self.vertices:
            string_representation += f'{item}: {self[item].get_neighbours()} \n'

        string_representation += '}'

        return 'Graph(' + string_representation + ')'



""" The file contains _DoubleNode and DoublyLinkedList class. Both of the classes have
ideas taken from University of Toronto's CSC111 course.
"""

from __future__ import annotations
from dataclasses import dataclass
from typing import Any, Optional, Sequence, Union, Callable
import math


@dataclass
class _DoubleNode:
    """A node in a doubly-linked list.

    The class is taken from University of Toronto's CSC111 course
    """
    item: Any
    prev: Optional[_DoubleNode] = None
    next: Optional[_DoubleNode] = None


class DoublyLinkedList:
    """The class represents a Doubly linked-list

    Idea for many of the methods is taken from the University of Toronto's CSC111 course
    """

    def __init__(self, items: Optional[Sequence] = None, first: Optional[_DoubleNode] = None,
                 last: Optional[_DoubleNode] = None) -> None:
        """Initialize a new doubly-linked list containing the given items.
        """

        if first and last:
            self._first = first
            self._last = last

            count = 0
            curr = self._first

            while curr is not None:
                count += 1
                curr = curr.next

            self._length = count
            return

        self._first = None
        self._last = None
        self._length = len(items) if items else 0

        if items:
            curr = None
            for item in items:
                if self._first is None:
                    self._first = _DoubleNode(item)
                    curr = self._first
                else:
                    new_node = _DoubleNode(item, curr)
                    curr.next = new_node
                    curr = curr.next
            self._last = curr

    def __len__(self) -> int:
        """Return the number of elements in this list.
        """
        return self._length

    def __getitem__(self, i: Union[int, slice]) -> Union[Any, DoublyLinkedList]:
        """Return the item stored at index i in this linked list.
        """
        if isinstance(i, int):
            if i < 0:
                i = len(self) + i

            if i >= len(self):
                raise IndexError

            elif i <= len(self) / 2:
                curr = self._first
                curr_index = 0
                while curr is not None:
                    if i == curr_index:
                        return curr.item
                    curr = curr.next
                    curr_index += 1

            elif i >= len(self) / 2:
                curr = self._last
                curr_index = len(self) - 1
                while curr is not None:
                    if i == curr_index:
                        return curr.item
                    curr = curr.prev
                    curr_index -= 1

        elif isinstance(i, slice):
            start = i.start
            stop = i.stop

            if start is None:
                start = 0
            if stop is None:
                stop = self._length

            if not (0 <= start <= stop <= len(self)):
                raise IndexError
            else:
                new_linked_list = DoublyLinkedList([])
                index = 0

                for item in self:
                    if start <= index < stop:
                        new_linked_list.append(item)
                    index += 1

                return new_linked_list

    def append(self, item: Any) -> None:
        """
        Append the item to the end of the list
        """
        new_node = _DoubleNode(item, self._last)

        if len(self) == 0:
            self._first = self._last = new_node
            self._length += 1
            return

        self._last.next = new_node
        self._last = new_node
        self._length += 1

    def to_list(self) -> list:
        """Return a built-in Python list containing the items of this linked list.

        The items in this linked list appear in the same order in the returned list.
        """
        items_so_far = []

        curr = self._first
        while curr is not None:
            items_so_far.append(curr.item)
            curr = curr.next

        return items_so_far

    def __contains__(self, item: Any) -> bool:
        """Return whether item is in this linked list."""
        if len(self) == 0:
            return False

        elif len(self) % 2 != 0:
            curr_from_start = self._first
            curr_from_end = self._last

            while curr_from_start != curr_from_end:
                if curr_from_end.item == item or curr_from_start.item == item:
                    return True
                curr_from_start = curr_from_start.next
                curr_from_end = curr_from_end.prev
            if curr_from_start and curr_from_start.item == item:
                return True
            return False
        else:
            if self._first.item == item:
                return True
            curr_from_start = self._first.next
            curr_from_end = self._last

            while curr_from_start != curr_from_end:
                if curr_from_end.item == item or curr_from_start.item == item:
                    return True
                curr_from_start = curr_from_start.next
                curr_from_end = curr_from_end.prev
            if curr_from_start and curr_from_start.item == item:
                return True
            return False

    def insert(self, i: int, item: Any) -> None:
        """
        Insert the item at index i in the list
        """
        if i < 0:
            i = len(self) + i

        if i > len(self):
            raise IndexError

        elif i <= len(self) / 2:
            curr = self._first
            curr_index = 0

            if i == 0:
                if self._first:
                    new_node = _DoubleNode(item, next=self._first)
                    self._first.prev = new_node
                    self._first = new_node
                else:
                    self._first = _DoubleNode(item)
                    self._last = self._first
                self._length += 1
                return

            while curr is not None:
                if curr_index == i - 1:
                    new_node = _DoubleNode(item, curr, curr.next)
                    if curr.next:
                        curr.next.prev = new_node
                    curr.next = new_node
                    self._length += 1
                    return
                curr = curr.next
                curr_index += 1

        elif i > len(self) / 2:
            curr = self._last
            curr_index = len(self) - 1

            while curr is not None:
                if curr_index == i - 1:
                    new_node = _DoubleNode(item, curr, curr.next)
                    if curr.next:
                        curr.next.prev = new_node
                    curr.next = new_node
                    self._length -= 1
                    return
                curr = curr.prev
                curr_index -= 1

    def pop(self, i: int) -> Any:
        if i < 0:
            i = len(self) + i

        if i >= len(self):
            raise IndexError

        elif i <= len(self) / 2:
            curr = self._first
            curr_index = 0

            if i == 0:
                item = self._first.item
                self._first = self._first.next

                if self._first:
                    self._first.prev = None

                self._length -= 1
                return item

            while curr is not None:
                if curr_index == i:
                    if curr.next:
                        curr.next.prev = curr.prev
                        if curr.prev:
                            curr.prev.next = curr.next
                    else:
                        if curr.prev:
                            curr.prev.next = None

                    self._length -= 1
                    return curr.item

                curr = curr.next
                curr_index += 1

        elif i > len(self) / 2:
            curr = self._last
            curr_index = len(self) - 1

            if i == curr_index:
                item = self._last.item

                if self._last.prev:
                    self._last.prev.next = None
                    self._last = self._last.prev

                self._length -= 1
                return item

            while curr is not None:
                if curr_index == i:
                    if curr.next:
                        curr.next.prev = curr.prev
                        curr.prev.next = curr.next
                    else:
                        curr.prev.next = None
                    self._length -= 1
                    return curr.item

                curr = curr.prev
                curr_index -= 1

    def __setitem__(self, i: int, item: Any) -> None:
        """Store item at index i in this list.
        """
        if i < 0:
            i = self._length + i

        if i < (self._length / 2):
            current_index = 0
            curr = self._first

            while curr is not None:
                if current_index == i:
                    curr.item = item

                curr = curr.next
                current_index += 1

        elif i < self._length:
            current_index = self._length - 1
            curr = self._last

            while curr is not None:
                if current_index == i:
                    curr.item = item

                curr = curr.prev
                current_index -= 1
        else:
            raise IndexError

    def index(self, item: Any) -> int:
        """Return the index of the first occurrence of the given item in this list.
        """

        index_so_far = 0
        curr = self._first

        while curr is not None:
            if curr.item == item:
                return index_so_far
            index_so_far += 1
            curr = curr.next
        raise ValueError

    def __str__(self) -> str:
        """Return a string representation of this list.
        """
        return '[' + ' <--> '.join([str(element) for element in self]) + ']'

    def __repr__(self) -> str:
        return f'DoubleLinkedList({self.__str__()})'

    def count(self, item: Any) -> int:
        """Return the number of times the given item occurs in this list.
        """
        count = 0

        if len(self) == 0:
            return count

        curr_from_start = self._first
        curr_from_end = self._last

        starting_index = 0

        if len(self) % 2 == 0:
            starting_index = 1
            curr_from_start = curr_from_start.next

            if self._first.item == item:
                count += 1

        ending_index = len(self) - 1

        while starting_index != ending_index:
            if curr_from_start.item == item:
                count += 1
            elif curr_from_end.item == item:
                count += 1

            starting_index += 1
            ending_index -= 1

            curr_from_start = curr_from_start.next
            curr_from_end = curr_from_end.prev

        if curr_from_start.item == curr_from_end.item == item:
            count += 1

        return count

    def copy(self) -> DoublyLinkedList:
        return DoublyLinkedList(first=self._first, last=self._last)

    def __add__(self, other: DoublyLinkedList) -> DoublyLinkedList:
        copy = self.copy()

        curr = other._first

        while curr is not None:
            copy.append(curr.item)
            curr = curr.next

        return copy

    def sort(self, reverse: Optional[bool] = False):
        lst = self.to_list()
        lst.sort(reverse=reverse)
        copy = DoublyLinkedList(lst)
        self._first = copy._first
        self._last = copy._last

    def remove(self, item: Any) -> None:
        """Remove the first occurrence
        """
        curr = self._first

        if not curr:
            raise ValueError

        elif curr.item == item:
            self._first = self._first.next

            if self._first:
                self._first.prev = None

            self._length -= 1

        else:
            while curr is not None:
                if curr.next and curr.next.item == item:
                    if curr.next.next:
                        curr.next.next.prev = curr.next
                    curr.next = curr.next.next

                    self._length -= 1
                    return
                curr = curr.next
            raise ValueError

    def __eq__(self, other: DoublyLinkedList) -> bool:
        """Return whether this list and the other list are equal.
        """
        curr1 = self._first
        curr2 = other._first
        are_equal = True

        while are_equal and curr1 is not None and curr2 is not None:
            if curr1.item != curr2.item:
                are_equal = False
            curr1 = curr1.next
            curr2 = curr2.next

        return are_equal

    def invert(self) -> None:
        """Invert the linked list
        """

        curr = self._first
        previous = None

        while curr:
            curr_value = curr.next
            curr.next = curr.prev
            curr.prev = curr_value
            
            previous = curr
            curr = curr_value

        self._last = self._first
        self._first = previous

    def extend(self, other: DoublyLinkedList) -> None:
        """Extend self by other linked list"""
        copy = self.copy()

        curr = other._first

        while curr is not None:
            copy.append(curr.item)
            curr = curr.next

        self._first = copy._first
        self._last = copy._last
        self._length = copy._length

    def map(self, key=Callable) -> DoublyLinkedList:
        """Return a mapped list to key"""
        lst = DoublyLinkedList()

        curr = self._first

        while curr is not None:
            lst.append(key(curr.item))
            curr = curr.next

        return lst

    def __abs__(self) -> DoublyLinkedList:
        return self.map(lambda x: abs(x))

    def abs(self) -> DoublyLinkedList:
        return self.map(lambda x: abs(x))

    def __floor__(self) -> DoublyLinkedList:
        return self.map(lambda x: math.floor(x))

    def floor(self) -> DoublyLinkedList:
        return self.map(lambda x: math.floor(x))

    def __ceil__(self) -> DoublyLinkedList:
        return self.map(lambda x: math.ceil(x))

    def ceil(self) -> DoublyLinkedList:
        return self.map(lambda x: math.ceil(x))



"""The file contains Queue class. The basic structure of the class id taken from the
University of Toronto's CSC110 course

This implementation of Queue is done on doubly linked lists.
"""
from __future__ import annotations

from structlinks.DataStructures.DoublyLinkedList.DoublyLinkedList import *
from structlinks.Algorithms.SortingAlgorithms.SortingAlgorithms import *

from typing import *


class Queue:
    """Represents a queue datatype"""
    def __init__(self, items: Optional[list] = None,
                 metric: Optional[Callable] = None,
                 limit: Optional[int] = None) -> None:

        self._items = DoublyLinkedList(items if items else [])
        self._metric = metric
        self._limit = limit

        if self._limit:
            while len(self._items) > self._limit:
                self._items.pop(0)

        if self._metric:
            quicksort(self._items, self._metric)

    @property
    def is_empty(self) -> bool:
        """Return whether this queue contains no items.
        """
        return len(self._items) == 0

    @property
    def is_filled(self) -> bool:
        """Return weather the queue us filled or not"""
        return self._limit and len(self._items) >= self._limit

    def copy(self) -> Queue:
        return Queue(items=self._items.to_list(), metric=self._metric, limit=self._limit)

    def __copy__(self) -> Queue:
        return Queue(items=self._items.to_list(), metric=self._metric, limit=self._limit)

    def to_list(self) -> list:
        """Return list representation of queue"""
        return self._items.to_list()

    def change_limit(self, limit: int) -> None:
        """Change the limit of the queue"""
        self._limit = limit

        if self._limit:
            while len(self._items) > self._limit:
                self._items.pop(0)

    def change_metric(self, metric: Callable) -> None:
        """Change the metric of the queue to <metric>"""
        self._metric = metric

        if self._metric:
            quicksort(self._items, self._metric)

    def enqueue(self, item) -> None:
        """put the item in the queue

        Raise an QueueLimitReachedError if this queue is full.
        """
        if not self._limit or self._limit > len(self._items):
            self._items.insert(0, item)

            if self._metric:
                quicksort(self._items, self._metric)
        else:
            raise QueueLimitReachedError

    def dequeue(self) -> Any:
        """Remove and return the item at the front of this queue.

        Raise an EmptyQueueError if this queue is empty."""
        if self.is_empty:
            raise EmptyQueueError
        else:
            return self._items.pop(len(self._items) - 1)

    def extend(self, other: Queue) -> None:
        """Extend self by other queue"""
        other_elements = other.to_list()

        while not self.is_filled and not other_elements == []:
            self.enqueue(other_elements.pop())

    def __len__(self) -> int:
        return len(self._items)

    def map(self, key: Callable) -> None:
        self._items = self._items.map(key)

    def update(self) -> None:
        """Update the order of the queue"""
        if self._metric:
            quicksort(self._items, self._metric)

    def display(self, key: Optional[Callable] = lambda x: x) -> None:
        string_so_far = ''

        temp = ''

        items = self._items.to_list()

        for index, item in enumerate(items):
            if index < len(items) - 1:
                temp += f' {key(item)} ->'
            else:
                temp += f' {key(item)}'

        string_so_far += 'Entry ' + '-' * len(temp) + '> Exit \n'

        string_so_far += '     ' + temp + '\n'

        string_so_far += '------' + '-' * len(temp) + '------'

        print(string_so_far)

    def __str__(self) -> str:
        return '[' + ' -> '.join([str(element) for element in self._items]) + ']'

    def __repr__(self) -> str:
        return f'Queue({self.__str__()})'


class EmptyQueueError(Exception):
    """Exception raised when calling dequeue on an empty queue.

    the exception is taken from University of Toronto's CSC110 course notes.
    """

    def __str__(self) -> str:
        """Return a string representation of this error."""
        return 'dequeue may not be called on an empty queue'


class QueueLimitReachedError(Exception):
    """Exception raised when calling dequeue on an filled queue.
    """

    def __str__(self) -> str:
        """Return a string representation of this error."""
        return 'dequeue may not be called on an filled queue'



"""
A re-implementation of several sorting algorithms, including:
    - Merge sort
    - Quick sort (In-place and out-of-place)
    - Selection sort
    - Insertion sort

Many of these sort are inspired by their implementations in the CSC111
University of Toronto Course.

Implementation Notes:
    - Non-mutating algorithms are explicitely prefixed with no_mut_
    - Every top level sorting algorithm (the main function of that algorithm)
        has a default value for key, (lambda x: x) which does not modify it's input
"""
from typing import Any, Callable, Union
from structlinks.DataStructures.LinkedList.LinkedList import LinkedList
from structlinks.DataStructures.DoublyLinkedList.DoublyLinkedList import DoublyLinkedList


########################################
# Merge sort (In-place)
########################################
def mergesort(lst: Union[list, LinkedList, DoublyLinkedList],
              key: Callable[[Any], Any] = (lambda x: x)) -> None:
    """
    Return a sorted list of the items in lst
    using the merge sort algorithm.

    This algorithm is heavily based off 'Approach 2' by Krikti,
    https://www.geeksforgeeks.org/in-place-merge-sort/
    """
    _in_place_mergesort(lst, 0, len(lst), key)


def _in_place_mergesort(lst: Union[list, LinkedList, DoublyLinkedList],
                        b: int, e: int, key: Callable[[Any], Any]) -> None:
    if e - b < 2:  # When there is only one element left in the list
        return
    else:
        m = ((e - b) // 2) + b  # Split the list in half

        # Sort each half individual
        _in_place_mergesort(lst, b, m, key)
        _in_place_mergesort(lst, m, e, key)

        # Merge and return the sorted half
        return _in_place_merge(lst, b, e, key)


def _in_place_merge(lst: Union[list, LinkedList, DoublyLinkedList],
                    b: int, e: int, key: Callable[[Any], Any]) -> None:
    """Return a single sorted list from two merged input lists."""

    # The initial gap between swappable elements
    gap = ((e - b + 1) + 1) // 2

    while gap > 0:
        for i in range(b, e - gap):
            j = i + gap
            if key(lst[j]) < key(lst[i]):
                lst[j], lst[i] = lst[i], lst[j]

        if gap <= 1:
            gap = 0
        else:
            # The +1 forces the ceiling of (gap / 2)
            gap = (gap + 1) // 2


########################################
# Merge sort (Non-mutating)
########################################
def no_mut_mergesort(lst: Union[list, LinkedList, DoublyLinkedList],
                     key: Callable[[Any], Any] = (lambda x: x)) -> \
        Union[list, LinkedList, DoublyLinkedList]:
    """
    Return a sorted list of the items in lst
    using the merge sort algorithm.
    """
    if isinstance(lst, LinkedList):
        return LinkedList(no_mut_mergesort(lst.to_list(), key))

    elif isinstance(lst, DoublyLinkedList):
        return DoublyLinkedList(no_mut_mergesort(lst.to_list(), key))

    else:
        if len(lst) < 2:  # When there is only one element left in the list
            return lst.copy()
        else:
            m = len(lst) // 2  # Split the list in half

            # Sort each half individual
            left = no_mut_mergesort(lst[:m], key)
            right = no_mut_mergesort(lst[m:], key)

            # Merge and return the sorted half
            return _no_mut_mergesort_merge(left, right, key)


def _no_mut_mergesort_merge(left: list, right: list, key: Callable[[Any], Any]) -> list:
    """Return a single sorted list from two merged input lists."""

    # Keep track of the current item being inspected in each list
    left_idx = 0
    right_idx = 0

    # Accumulator for the new list
    sorted_so_far = []
    # Loop until we reach the end of one list
    while left_idx < len(left) and right_idx < len(right):
        # assert sorted_so_far == sorted(left[:left_idx] + right[:right_idx], key=key)

        if key(left[left_idx]) <= key(right[right_idx]):
            sorted_so_far.append(left[left_idx])
            left_idx -= - 1
        else:
            sorted_so_far.append(right[right_idx])
            right_idx -= - 1

    assert left_idx == len(left) or right_idx == len(right)

    # Add the remaining element in one of the lists to the sorted list
    if left_idx < len(left):
        return sorted_so_far + left[left_idx:]
    elif right_idx < len(right):
        return sorted_so_far + right[right_idx:]


########################################
# Quick sort (In Place)
########################################
def quicksort(lst: Union[list, LinkedList, DoublyLinkedList],
              key: Callable[[Any], Any] = (lambda x: x)) -> None:
    """An in-place implementation of the quick-sort algorithm.
    """

    _in_place_quicksort(lst, 0, len(lst), key)


def _in_place_quicksort(lst: Union[list, LinkedList, DoublyLinkedList],
                        b: int, e: int, key: Callable[[Any], Any]) -> None:
    """The main helper method of the in-place quicksort algorithm.
    """
    if e - b < 2:  # If there is only one element left in the list
        pass
    else:
        # Partition the list into entries smaller than the pivot
        # and entries larger than the pivot
        pivot = _in_place_partition(lst, b, e, key)

        # Sort both sides of the pivot
        _in_place_quicksort(lst, 0, pivot, key)
        _in_place_quicksort(lst, pivot + 1, e, key)


def _in_place_partition(lst: Union[list, LinkedList, DoublyLinkedList],
                        b: int, e: int, key: Callable[[Any], Any]) -> int:
    """Partition the input list between indexes b and e using the pivot
    at index b (the pivot is lst[b]).

    Return the index of the pivot after the inplace operations are complete.
    """

    pivot = key(lst[b])
    left_idx = b + 1
    right_idx = e

    # Close in on values from the smallest and largest sides
    while left_idx != right_idx:
        # If item at left index is smaller than the pivot, then move out index
        # up one, confirming that fact
        if key(lst[left_idx]) <= pivot:
            left_idx -= - 1
        else:
            # Otherwise, swap the entry to the end and extend the region that
            # is larger than the pivot by moving the boundry index closer to
            # the center of the list
            lst[right_idx - 1], lst[left_idx] = lst[left_idx], lst[right_idx - 1]
            right_idx += - 1

    # Swap the pivot to the middle of the smaller and larger groups
    lst[b], lst[left_idx - 1] = lst[left_idx - 1], lst[b]

    return left_idx - 1


########################################
# Quick sort (Non-mutating)
########################################
def no_mut_quicksort(lst: Union[list, LinkedList, DoublyLinkedList],
                     key: Callable[[Any], Any] = (lambda x: x)) -> \
        Union[list, LinkedList, DoublyLinkedList]:
    """Return a sorted list with the elements of lst using the quicksort
    algorithm without in-place optimizations: each recursive call creates a
    new list to be sorted.
    """
    if len(lst) < 2:
        return lst
    else:
        pivot = lst[0]
        left, right = _no_mut_partition(lst[1:], pivot, key)  # We remove the pivot

        middle = [lst[0]]

        if isinstance(lst, LinkedList):
            middle = LinkedList(middle)

        if isinstance(lst, DoublyLinkedList):
            middle = DoublyLinkedList(middle)

        return no_mut_quicksort(left, key) + middle + no_mut_quicksort(right, key)


def _no_mut_partition(lst: Union[list, LinkedList, DoublyLinkedList],
                      pivot: Any, key: Callable[[Any], Any]) -> \
        tuple[Union[list, LinkedList, DoublyLinkedList],
              Union[list, LinkedList, DoublyLinkedList]]:
    """Partition the list lst relative to the given pivot.

    The first index of the tuple is a list of all items smaller than the pivot,
    the second index is a list of all items greater than the pivot
    """

    smaller_lst = []
    larger_lst = []

    if isinstance(lst, LinkedList):
        smaller_lst = LinkedList()
        larger_lst = LinkedList()

    if isinstance(lst, DoublyLinkedList):
        smaller_lst = DoublyLinkedList()
        larger_lst = DoublyLinkedList()

    for item in lst:
        if key(item) <= key(pivot):
            smaller_lst.append(item)
        else:
            larger_lst.append(item)

    return (smaller_lst, larger_lst)


########################################
# Selection Sort
########################################
def selection_sort(lst: Union[list, LinkedList, DoublyLinkedList],
                   key: Callable[[Any], Any] = (lambda x: x)) -> None:
    """An in-place (mutating) implementation of the selection sort algorithm.
    """

    # Build a sorted list from the smallest elements in the unsorted portion of
    # the list
    for idx in range(len(lst)):
        min_index = _smallest_index(lst, idx, key)
        lst[idx], lst[min_index] = lst[min_index], lst[idx]


def _smallest_index(lst: Union[list, LinkedList, DoublyLinkedList],
                    i: int, key: Callable[[Any], Any]) -> int:
    """Return the index of the smallest item in the sublist lst[i:]
    """

    # Extract the smallest item
    smallest_so_far = i
    for idx in range(i + 1, len(lst)):
        if key(lst[smallest_so_far]) > key(lst[idx]):
            smallest_so_far = idx

    return smallest_so_far


########################################
# Selection Sort
########################################
def insertion_sort(lst: Union[list, LinkedList, DoublyLinkedList],
                   key: Callable[[Any], Any] = (lambda x: x)) -> None:
    """An in-place (mutating) implementation of the insertion sort algorithm.
    """
    for idx in range(0, len(lst)):
        _insert(lst, idx, key)


def _insert(lst: Union[list, LinkedList, DoublyLinkedList],
            i: int, key: Callable[[Any], Any]) -> None:
    """Move lst[i] so that lst[:i + 1] is sorted.
    """
    idx = i

    # "Shift" the item at index to the left until the item is greater than the item
    # in the index before it
    while idx > 0 and key(lst[idx]) < key(lst[idx - 1]):
        lst[idx], lst[idx - 1] = lst[idx - 1], lst[idx]

        idx -= 1



"""This file contains the Binary search tree class. SOme of the methods of the class are
taken from University of Toronto's CSC111 course

The printing of the branched binary search tree is taken from StackOverflow, the link
is the the readme.md file.
"""

from __future__ import annotations
from typing import Optional, Any, Callable
import math
import copy as cpy


class BinarySearchTree:
    """The class represents a Binary search tree

    Some of the methods are taken from the University of Toronto's CSC111 course
    """
    def __init__(self, root: Optional[Any]) -> None:

        if root is None:
            self._root = None
            self._left = None
            self._right = None
        else:
            self._root = root
            self._left = BinarySearchTree(None)
            self._right = BinarySearchTree(None)

    def set_left_to(self, tree: BinarySearchTree) -> None:
        if isinstance(tree, BinarySearchTree):
            self._left = tree
        else:
            raise TypeError

    def set_right_to(self, tree: BinarySearchTree) -> None:
        if isinstance(tree, BinarySearchTree):
            self._right = tree
        else:
            raise TypeError

    @property
    def root(self) -> Any:
        """Root of the BST"""
        return cpy.copy(self._root)

    @property
    def left(self) -> Any:
        """Return copy of left child of the BST"""
        return self._left.copy()

    @property
    def right(self) -> Any:
        """Return copy of right child of the BST"""
        return self._right.copy()

    @property
    def is_balanced(self) -> bool:

        if self.is_empty():
            return True

        if abs(self._left.height - self._right.height) <= 1 and \
                self._left.is_balanced and self._right.is_balanced:
            return True

        return False

    @property
    def height(self) -> int:
        """Return the height of the tree"""
        if self.is_empty():
            return 0
        else:
            return 1 + max(self._left.height, self._right.height)

    def is_empty(self) -> bool:
        """Return whether this BST is empty.
        """
        return self._root is None

    def __contains__(self, item: Any) -> bool:
        """Return whether <item> is in this BST.
        """
        if self.is_empty():
            return False
        elif item == self._root:
            return True
        elif item < self._root:
            return self._left.__contains__(item)
        else:
            return self._right.__contains__(item)

    def __copy__(self) -> BinarySearchTree:
        """Return a copy of the BST"""
        tree = BinarySearchTree(self._root)
        tree.set_right_to(self._right)
        tree.set_left_to(self._left)
        return tree

    def copy(self) -> BinarySearchTree:
        """Return a copy of the BST"""
        if self.is_empty():
            return BinarySearchTree(self.root)
        else:
            tree = BinarySearchTree(self.root)
            tree.set_right_to(self._right.copy())
            tree.set_left_to(self._left.copy())
            return tree

    def __abs__(self) -> BinarySearchTree:
        """Return BST similar to self, but with abs values"""
        return self.map(lambda x: abs(x))

    def abs(self) -> BinarySearchTree:
        """Return BST similar to self, but with abs values"""
        return self.map(lambda x: abs(x))

    def __eq__(self, other: BinarySearchTree) -> bool:
        """Check weather two BST's ate same or not"""
        return self.to_list() == other.to_list()

    def floor(self) -> BinarySearchTree:
        """Return BST similar to self, but with floor values"""
        return self.map(lambda x: math.floor(x))

    def ceil(self) -> BinarySearchTree:
        """Return BST similar to self, but with ceil values"""
        return self.map(lambda x: math.ceil(x))

    def __floor__(self) -> BinarySearchTree:
        """Return BST similar to self, but with floor values"""
        return self.map(lambda x: math.floor(x))

    def __ceil__(self) -> BinarySearchTree:
        """Return BST similar to self, but with ceil values"""
        return self.map(lambda x: math.ceil(x))

    def map(self, key=Callable) -> BinarySearchTree:
        """Map the BST to key and return mapped BST"""
        copy = self.copy()

        if not copy.is_empty():
            copy._root = key(self._root)
            copy.set_left_to(copy._left.map(key))
            copy.set_right_to(self._right.map(key))

        return copy

    def apply(self, key=Callable) -> None:
        """Map the BST to key"""

        if not self.is_empty():
            self._root = key(self._root)
            self._left.apply(key)
            self._right.apply(key)

    def __str__(self) -> str:
        """Return a string representation of this BST.
        """
        return self.display_branched()

    def __repr__(self) -> str:
        return 'BinarySearchTree( \n' + self.display_branched() + ')'

    def display(self, indented: Optional[bool] = False):
        """Print a string representation of this BST"""
        if indented:
            print(self.display_indented())
        else:
            print(self.display_branched())

    def display_indented(self) -> str:
        """Return indented string representation of BST"""
        return self._str_indented(0)

    def _str_indented(self, depth: int) -> str:
        """Return an indented string representation of this BST.
        """
        if self.is_empty():
            return ''
        else:
            return (
                    depth * '  ' + '|->' + f'{self._root}\n'
                    + self._left._str_indented(depth + 1)
                    + self._right._str_indented(depth + 1)
            )

    def display_branched(self) -> str:
        string_so_far = ''
        lines, *_ = self._display_aux()
        for line in lines:
            string_so_far += line + '\n'

        return string_so_far

    def _display_aux(self) -> Any:

        if self._right.is_empty() and self._left.is_empty():
            line = '%s' % self._root
            width = len(line)
            height = 1
            middle = width // 2

            return [line], width, height, middle

        if self._right.is_empty():
            lines, n, p, x = self._left._display_aux()

            s = '%s' % self._root
            u = len(s)
            first_line = (x + 1) * ' ' + (n - x - 1) * '_' + s
            second_line = x * ' ' + '/' + (n - x - 1 + u) * ' '
            shifted_lines = [line + u * ' ' for line in lines]

            return [first_line, second_line] + shifted_lines, n + u, p + 2, n + u // 2

        if self._left.is_empty():
            lines, n, p, x = self._right._display_aux()

            s = '%s' % self._root
            u = len(s)
            first_line = s + x * '_' + (n - x) * ' '
            second_line = (u + x) * ' ' + '\\' + (n - x - 1) * ' '
            shifted_lines = [u * ' ' + line for line in lines]

            return [first_line, second_line] + shifted_lines, n + u, p + 2, u // 2

        left, n, p, x = self._left._display_aux()
        right, m, q, y = self._right._display_aux()

        s = '%s' % self._root
        u = len(s)
        first_line = (x + 1) * ' ' + (n - x - 1) * '_' + s + y * '_' + (m - y) * ' '
        second_line = x * ' ' + '/' + (n - x - 1 + u + y) * ' ' + '\\' + (m - y - 1) * ' '
        if p < q:
            left += [n * ' '] * (q - p)
        elif q < p:
            right += [m * ' '] * (p - q)
        zipped_lines = zip(left, right)
        lines = [first_line, second_line] + [a + u * ' ' + b for a, b in zipped_lines]

        return lines, n + m + u, max(p, q) + 2, n + u // 2

    def maximum(self) -> Optional[int]:
        """Return the maximum number in this BST, or None if this BST is empty.
        """
        if self.is_empty():
            return None
        else:
            curr_max = self._root

            if self._right._root and curr_max < self._right._root:
                curr_max = self._right.maximum()

            return curr_max

    def minimum(self) -> Optional[Any]:
        """Return the minimum number in this BST, or None if this BST is empty.
        """
        if self.is_empty():
            return None
        else:
            curr_min = self._root

            if self._right._root and curr_min > self._left._root:
                curr_min = self._left.maximum()

            return curr_min

    def items(self) -> list:
        """Return all of the items in the BST in sorted order.
        """
        if self.is_empty():
            return []
        else:
            return self._left.items() + [self._root] + self._right.items()

    def insert(self, item: Any) -> None:
        """Insert item into the list
        """
        if self.is_empty():
            self._root = item
            self._left = BinarySearchTree(None)
            self._right = BinarySearchTree(None)
        elif item > self._root:
            self._right.insert(item)
        elif item < self._root:
            self._left.insert(item)
        else:
            return

    def remove(self, item) -> bool:
        """Remove the item from the tree
        """
        if self.is_empty():
            return False
        elif self._root == item:
            self._delete_root()
        else:
            if item > self._root:
                return self._right.remove(item)
            else:
                return self._left.remove(item)

    def _delete_root(self) -> None:

        if self._left.is_empty() and self._right.is_empty():

            self._root = None
            self._right._root = None
            self._left._root = None

        elif self._left.is_empty():

            self._root, self._left, self._right = self._right._root, \
                                                  self._right._left, \
                                                  self._right._right

        elif self._right.is_empty():

            self._root, self._left, self._right = self._left._root, \
                                                  self._left._left, \
                                                  self._left._right
        else:
            self._root = self._extract_max()

    def _extract_max(self) -> Any:

        if self._right._root is None:  # if this is the case then _root is the largest number
            max_item = self._root
            self._root, self._right._root, self._left._root = None, None, None
            return max_item
        else:
            return self._right._extract_max()

    def invert(self) -> None:
        """Invert the BST"""
        if not self.is_empty():

            if not self._left.is_empty():
                self._left.invert()
            if not self._right.is_empty():
                self._right.invert()

            temp = self._left
            self._left = self._right
            self._right = temp

    @staticmethod
    def create_tree(lst: list) -> BinarySearchTree:
        """Return a balanced binary search tree"""
        lst = list(set(lst))

        if len(lst) == 1:
            return BinarySearchTree(lst[0])

        elif len(lst) > 1:

            lst = sorted(lst)

            middle = len(lst) // 2

            left = BinarySearchTree.create_tree(lst[:middle])

            right = None

            if middle + 1 < len(lst):
                right = BinarySearchTree.create_tree(lst[(middle + 1):])

            tree = BinarySearchTree(lst[middle])

            tree._left = left

            if right:
                tree._right = right

            return tree

    def to_list(self) -> list:
        """Return all of the items in the BST in sorted order.
        """
        if self.is_empty():
            return []
        else:
            return self._left.to_list() + [self._root] + self._right.to_list()

    def balance(self) -> None:
        """Convert the BST to a balanced BST"""
        tree = BinarySearchTree.create_tree(self.to_list())

        self._root = tree._root
        self._left = tree._left
        self._right = tree._right



import setuptools

with open('structlinks/README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='structlinks',
    version='1.1.4',
    author='Eeshan Narula',
    author_email='eeshannarula29@gmail.com',
    description='Easily access and visualize Data structures',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/eeshannarula29/structlinks',
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9.0'
)



from __future__ import annotations
from typing import Optional, Callable, Any 

from math import floor, ceil, log2

class PriorityFunction:
    def __init__(self, comaprison_function: Callable) -> None:
        self._function = comaprison_function
    
    def compare(self, parent, child) -> bool:
        return self._function(parent, child)
            
MAXPRIORITY = PriorityFunction(lambda compare_to, compare_from : compare_to >= compare_from)
MINPRIORITY = PriorityFunction(lambda compare_to, compare_from : compare_to <= compare_from)

class Heap:
    
    def __init__(self, prioritizer: PriorityFunction, inital_heep: Optional[list[Any]] = []) -> None:
        self._tree = inital_heep
        self.compare_function = prioritizer
        
        self._heapify()
        
    def _get_children(self, index) -> list[dict]:
        children = []
        
        if 2 * index + 1 < len(self._tree):
            children.append({
                'index': 2 * index + 1,
                'value': self._tree[2 * index + 1]
            })
        
        if 2 * index + 2 < len(self._tree):
            children.append({
                'index': 2 * index + 2,
                'value': self._tree[2 * index + 2]
            })    
        
        return children
    
    @property
    def size(self) -> int:
        return len(self._tree)
    
    @property
    def height(self) -> int:
        return ceil(log2(self.size + 1)) - 1
    
    def to_list(self) -> list:
        return self._tree.copy()
    
    def _is_leaf(self, index) -> bool:
        return  2 * index + 1 >= len(self._tree) and 2 * index + 2 >= len(self._tree)
    
    def _is_root(self, index) -> bool:
        return floor((index - 1)/2) < 0
    
    def _get_parent(self, index) -> dict:
        return {
            'index': floor((index - 1)/2),
            'value': self._tree[floor((index - 1)/2)] if not self._is_root(index) else None
        } 
        
    def _heapify_update(self, root_index) -> None:
        
        largest = root_index
        
        for child in self._get_children(root_index):
            if not self.compare_function.compare(self._tree[largest], child['value']): 
                largest = child['index']
        
        if not largest == root_index:
            self._tree[root_index], self._tree[largest] = self._tree[largest], self._tree[root_index]
            self._heapify_update(largest)   
    
    def _heapify(self) -> None:
        
        for index in range(len(self._tree) - 1, -1, -1):
            self._heapify_update(index)    
    
    def push(self, element) -> None:
        self._tree.append(element)
        
        child_index = len(self._tree) - 1
        parent_index = self._get_parent(child_index)
        
        while parent_index['value'] and not self.compare_function.compare(parent_index['value'], self._tree[child_index]):
            
            self._tree[child_index], self._tree[parent_index['index']] = self._tree[parent_index['index']], self._tree[child_index]
            child_index, parent_index = parent_index['index'], self._get_parent(parent_index['index'])

    def _pop_update(self, root_index) -> None:
            largest = root_index
            
            for child in self._get_children(root_index):
                if not self.compare_function.compare(self._tree[largest], child['value']): 
                    largest = child['index']
            
            if not largest == root_index:
                self._tree[root_index], self._tree[largest] = self._tree[largest], self._tree[root_index]
                self._pop_update(largest)   
            
    def pop(self) -> Any:
            self._tree[0], self._tree[len(self._tree) -1] = self._tree[len(self._tree) -1], self._tree[0]
            popped_element = self._tree.pop()
            self._pop_update(0)
            
            return popped_element
            
    def __str__(self) -> str:
        return str(self.display() if self.size > 0 else 'Empty Heap')
    
    def __repr__(self) -> str:
        return str(f'Heep(\n{self.display() if self.size > 0 else "Empty Heap"}\n)')     
    
    def display(self, listed: bool = False):
        return self._display_branched() if not listed else self._display_listed()
    
    def _display_listed(self) -> str:
        return str(self._tree)
    
    def _display_branched(self) -> str:
        string_so_far = ''
        lines, *_ = self._display_aux(0)
        for line in lines:
            string_so_far += line + '\n'

        return string_so_far

    def _display_aux(self, root_index) -> Any:
        
            children = self._get_children(root_index)
            
            if children == []:
                line = '%s' % self._tree[root_index]
                width = len(line)
                height = 1
                middle = width // 2
                
                return [line], width, height, middle

            if len(children) == 1:
                lines, n, p, x = self._display_aux(children[0]['index'])

                s = '%s' % self._tree[root_index]
                u = len(s)
                first_line = (x + 1) * ' ' + (n - x - 1) * '_' + s
                second_line = x * ' ' + '/' + (n - x - 1 + u) * ' '
                shifted_lines = [line + u * ' ' for line in lines]

                return [first_line, second_line] + shifted_lines, n + u, p + 2, n + u // 2  

            if len(children) == 2:
                left, n, p, x = self._display_aux(children[0]['index'])
                right, m, q, y = self._display_aux(children[1]['index'])

                s = '%s' % self._tree[root_index]
                u = len(s)
                first_line = (x + 1) * ' ' + (n - x - 1) * '_' + s + y * '_' + (m - y) * ' '
                second_line = x * ' ' + '/' + (n - x - 1 + u + y) * ' ' + '\\' + (m - y - 1) * ' '
                if p < q:
                    left += [n * ' '] * (q - p)
                elif q < p:
                    right += [m * ' '] * (p - q)
                zipped_lines = zip(left, right)
                lines = [first_line, second_line] + [a + u * ' ' + b for a, b in zipped_lines]

                return lines, n + m + u, max(p, q) + 2, n + u // 2                         

class MinHeap(Heap):
    def __init__(self, inital_heep: Optional[list[Any]] = []) -> None:
        super().__init__(inital_heep=inital_heep, prioritizer=MINPRIORITY)    

class MaxHeap(Heap):
    def __init__(self, inital_heep: Optional[list[Any]] = []) -> None:
        super().__init__(inital_heep=inital_heep, prioritizer=MAXPRIORITY)                


"""This file contains the Node and Linked-list class. Both of these classes have
many methods which are taken form the University of Toronto's CSC111 course. The
"""

from __future__ import annotations
from dataclasses import dataclass
from typing import Any, Optional, Sequence, Union, Callable
import math


@dataclass
class _Node:
    """A node in a linked list.

    The class it take from University of Toronto CSC111 course
    """
    item: Any
    next: Optional[_Node] = None


class LinkedList:
    """The class represents a Linked lists

    Many of the methods in the class are taken from University of Toronto CSC111 course
    """

    def __init__(self, items: Optional[Sequence] = None, first: Optional[_Node] = None) -> None:
        """Initialize a linked list.
        """
        self._first = first
        self._length = 0

        if items:
            for item in items:
                self.append(item)

    def append(self, item: Any) -> None:
        """Append item to the end of this list.
        """
        new_node = _Node(item)

        if self._first is None:
            self._first = new_node
        else:
            curr = self._first
            while curr.next is not None:
                curr = curr.next

            curr.next = new_node
        self._length += 1

    def __contains__(self, item: Any) -> bool:
        """Return whether item is in this linked list.
        """
        curr = self._first

        while curr is not None:
            if curr.item == item:
                return True

            curr = curr.next

        return False

    def __getitem__(self, i: Union[int, slice]) -> Union[Any, LinkedList]:
        """Return the item stored at index i in this linked list.
        """

        if isinstance(i, int):

            if i < 0:
                i = self._length + i

            curr = self._first
            curr_index = 0

            while curr is not None and curr_index != i:
                curr = curr.next
                curr_index += 1

            if curr is None:
                raise IndexError
            else:
                return curr.item
        else:
            start = i.start
            stop = i.stop

            if start is None:
                start = 0
            if stop is None:
                stop = self._length

            if not (0 <= start <= stop <= len(self)):
                raise IndexError
            else:
                new_linked_list = LinkedList([])
                index = 0

                for item in self:
                    if start <= index < stop:
                        new_linked_list.append(item)
                    elif index > stop:
                        break
                    index += 1

                return new_linked_list

    def __len__(self) -> int:
        """Return the number of elements in this linked list.
        """
        return self._length

    def __eq__(self, other: LinkedList) -> bool:
        """Return whether this list and the other list are equal.
        """
        curr1 = self._first
        curr2 = other._first
        are_equal = True

        while are_equal and curr1 is not None and curr2 is not None:
            if curr1.item != curr2.item:
                are_equal = False
            curr1 = curr1.next
            curr2 = curr2.next

        return are_equal

    def __str__(self) -> str:
        """Return a string representation of this list.
        """
        return '[' + ' -> '.join([str(element) for element in self]) + ']'

    def __repr__(self) -> str:
        return f'LinkedList({self.__str__()})'

    def insert(self, i: int, item: Any) -> None:
        """Insert the given item at index i in this list.
        """
        if i < 0:
            i = self._length + i

        if i == 0:
            next_node = self._first
            self._first = _Node(item, next_node)
            self._length += 1
        else:
            curr = self._first
            curr_index = 0

            while curr is not None and curr_index != i - 1:
                curr_index += 1
                curr = curr.next

            if curr is None:
                raise IndexError
            else:
                new_node = _Node(item, curr.next)
                curr.next = new_node
                self._length += 1

    def pop(self, i: int) -> Any:
        """Remove and return the item at index i.
        """
        if i < 0:
            i = self._length + i

        if i == 0:
            if self._first:
                item = self._first.item
                self._first = self._first.next
                self._length -= 1
                return item
            else:
                raise IndexError
        else:
            curr = self._first
            curr_index = 0

            while curr is not None:
                if curr_index == i - 1:
                    if curr.next is None:
                        raise IndexError
                    else:
                        item = curr.next.item
                        curr.next = curr.next.next
                        self._length -= 1
                        return item

                curr = curr.next
                curr_index += 1
            raise IndexError

    def remove(self, item: Any) -> None:
        """Remove the first occurrence of item from the list.
        """
        curr = self._first

        if not curr:
            raise ValueError

        elif curr.item == item:
            self._first = self._first.next
            self._length -= 1

        else:
            while curr is not None:
                if curr.next and curr.next.item == item:
                    curr.next = curr.next.next
                    self._length -= 1
                    return
                curr = curr.next
            raise ValueError

    def count(self, item: Any) -> int:
        """Return the number of times the given item occurs in this list.
        """
        curr = self._first
        count = 0

        while curr is not None:
            if curr.item == item:
                count += 1
            curr = curr.next

        return count

    def index(self, item: Any) -> int:
        """Return the index of the first occurrence of the given item in this list.
        """
        index_so_far = 0
        curr = self._first

        while curr is not None:
            if curr.item == item:
                return index_so_far
            index_so_far += 1
            curr = curr.next
        raise ValueError

    def __setitem__(self, i: int, item: Any) -> None:
        """Store item at index i in this list.
        """
        if i < 0:
            i = self._length + i

        curr = self._first
        index_so_far = 0

        while curr is not None:
            if index_so_far == i:
                curr.item = item
                break
            index_so_far += 1
            curr = curr.next
        if curr is None:
            raise IndexError

    def invert(self) -> None:
        """Invert the linked list
        """

        curr = self._first
        previous = None

        while curr:
            curr_value = curr.next
            curr.next = previous

            previous = curr
            curr = curr_value

        self._first = previous

    def to_list(self) -> list:
        """Return a built-in Python list containing the items of this linked list.
        """
        items_so_far = []

        curr = self._first
        while curr is not None:
            items_so_far.append(curr.item)
            curr = curr.next

        return items_so_far

    def copy(self) -> LinkedList:
        return LinkedList(first=self._first)

    def extend(self, other: LinkedList) -> None:
        """Extend self by other linked list"""
        copy = self.copy()

        curr = other._first

        while curr is not None:
            copy.append(curr.item)
            curr = curr.next

        self._first = copy._first

    def __add__(self, other: LinkedList) -> LinkedList:
        copy = self.copy()

        curr = other._first

        while curr is not None:
            copy.append(curr.item)
            curr = curr.next

        return copy

    def sort(self, reverse: Optional[bool] = False):
        lst = self.to_list()
        lst.sort(reverse=reverse)
        copy = LinkedList(lst)
        self._first = copy._first

    def map(self, key=Callable) -> LinkedList:
        """Return a mapped list to key"""
        lst = LinkedList()

        curr = self._first

        while curr is not None:
            lst.append(key(curr.item))
            curr = curr.next

        return lst

    def __abs__(self) -> LinkedList:
        return self.map(lambda x: abs(x))

    def abs(self) -> LinkedList:
        return self.map(lambda x: abs(x))

    def __floor__(self) -> LinkedList:
        return self.map(lambda x: math.floor(x))

    def floor(self) -> LinkedList:
        return self.map(lambda x: math.floor(x))

    def __ceil__(self) -> LinkedList:
        return self.map(lambda x: math.ceil(x))

    def ceil(self) -> LinkedList:
        return self.map(lambda x: math.ceil(x))



from structlinks.DataStructures.BinarySearchTree.BinarySearchTree import BinarySearchTree
from structlinks.DataStructures.DoublyLinkedList.DoublyLinkedList import DoublyLinkedList
from structlinks.DataStructures.Graph.Graph import Graph
from structlinks.DataStructures.LinkedList.LinkedList import LinkedList
from structlinks.DataStructures.Matrix.Matrix import Matrix
from structlinks.DataStructures.Queue.Queue import Queue
from structlinks.DataStructures.Heap.Heap import Heap, MaxHeap, MinHeap, PriorityFunction